/********************************************************************
 File: usb.c

 Description:
 This file contains the USB stack for the USB2FIR firmware project.

 Authors and Copyright:
 (c) 2018, Thomas Fischl (http://www.fischl.de/contact.html)

 Device: PIC16LF1455
 Compiler: Microchip MPLAB XC8 C Compiler V1.44

 License:
 This file is part of the USBtin firmware project and is copyrighted by the
 authors listed above. It is free for private or educational non-commercial use
 (for a commercial license please contact the authors). It may not be
 redistributed without the prior written consent of the authors.
 
 It is provided "as is" without warranty of any kind, either express or implied,
 including without limitation any implied warranties of condition, uninterrupted
 use, merchantability, fitness for a particular purpose, or non-infringement. In
 no event shall the authors be liable for any direct or indirect damages arising
 in any way out of the use of it.
 
 ********************************************************************/

#include <xc.h>
#include "usb_descr.h"
#include "usb.h"

volatile EndpointType ep[EP_MAX] @ 0x2000;
volatile unsigned char ep0out_buffer[EP_BUFFERSIZE] @ 0x2020;
volatile unsigned char ep0in_buffer[EP_BUFFERSIZE] @ 0x2028;
volatile unsigned char ep1in_buffer[EP_BULK_BUFFERSIZE] @ 0x2100;

unsigned short usb_sendleft = 0;
unsigned char usb_sendpending = 0;
unsigned char usb_usertransfer = 0;
unsigned short usb_usertransfer_length = 0;

const unsigned char * usb_sendbuffer;
unsigned configured = 0;
unsigned char usb_config = 0;
unsigned char usb_setaddress = 0;
unsigned char usb_ep0status[2];

/**
 * @brief Initialize USB stack
 */
void usb_init() {
    ep[0].out.stat = 0x80;
    ep[0].out.cnt = EP_BUFFERSIZE;
    ep[0].out.adrl = 0x20;
    ep[0].out.adrh = 0x20;

    ep[0].in.stat = 0;
    ep[0].in.cnt = EP_BUFFERSIZE;
    ep[0].in.adrl = 0x28;
    ep[0].in.adrh = 0x20;

    ep[1].in.stat = 0x40;
    ep[1].in.cnt = EP_BULK_BUFFERSIZE;
    ep[1].in.adrl = 0x00;
    ep[1].in.adrh = 0x21;
    
    UEP0 = 0x16; // enable IN, enable OUT, enable CONTROL, enable handshake
    UEP1 = 0x1A; // enable IN, disable OUT, disable CONTROL, enable handshake

    UCFG = 0x14; // enable pullup and full-speed
    UCON = 0x08; // enable usb module

    usb_ep0status[0] = 0;
    usb_ep0status[1] = 0;
}

/**
 * @brief Shutdown USB stack
 */
void usb_shutdown() {
    UCON = 0x00;
    UCFG = 0x00;
}

/**
 * @brief Determine if serial number is available
 * @return 0 if no number available
 */
unsigned char usb_serialNumberAvailable() {
    return ((usb_string_serial[0] == USB_STRING_SERIALNUMBER_SIZE) && (usb_string_serial[1] == 0x03));
}

/**
 * @brief Process pending send activity
 */
void usb_sendProcess() {
    if (!usb_sendpending) return;

    unsigned short length = usb_sendleft;
    if (length < EP_BUFFERSIZE) {
        // last packet
        usb_sendpending = 1; 
    } else {
        length = EP_BUFFERSIZE;
    }

    unsigned char i;
    for (i = 0; i < length; i++) {

        if ((usb_sendbuffer == usb_dev_desc + USB_DEV_DESC_SERIALNUMBER_OFFSET) && usb_serialNumberAvailable()) ep0in_buffer[i] = USB_STRING_SERIALNUMBER_INDEX;
        else ep0in_buffer[i] = *usb_sendbuffer;

        usb_sendbuffer++;
        usb_sendleft--;
    }

    ep[0].in.cnt = length;
    if (ep[0].in.stat & 0x40)
        ep[0].in.stat = 0x88;
    else
        ep[0].in.stat = 0xC8;
}

/**
 * @brief Load given descriptor buffer into send buffer
 *
 * @param descbuffer Pointer to descriptor
 * @param size Size of descriptor
 * @param length Count of bytes to send
 * @return 0 on error, 1 on success
 */
unsigned char usb_loadDescriptor(const unsigned char * descbuffer, unsigned short size, unsigned short length) {

    if (length > size)
        length = size;

    usb_sendleft = length;
    usb_sendbuffer = descbuffer;
    usb_sendpending = 1;

    usb_sendProcess();
    return length;
}

/**
 * @brief Handle descriptor requests
 *
 * @param type Type of descriptor
 * @param index Index of descriptor
 * @param length Requested length
 * @return 0 on error, 1 on success
 */
unsigned char usb_handleDescriptorRequest(unsigned char type, unsigned char index, unsigned short length) {

    switch (type) {
        case DESCR_DEVICE:
            return usb_loadDescriptor(usb_dev_desc, sizeof (usb_dev_desc), length);
        case DESCR_CONFIG:
            return usb_loadDescriptor(usb_config_desc, sizeof (usb_config_desc), length);
        case DESCR_STRING:
            switch (index) {
                case 0: return usb_loadDescriptor(usb_string_language, sizeof (usb_string_language), length);
                case 1: return usb_loadDescriptor(usb_string_manuf, sizeof (usb_string_manuf), length);
                case 2: return usb_loadDescriptor(usb_string_product, sizeof (usb_string_product), length);
                case 3: return usb_loadDescriptor(usb_string_serial, USB_STRING_SERIALNUMBER_SIZE, length);
            }
    }

    return 0;
}

/**
 * @brief Do IN transfer
 */
void ut_in() {

    unsigned short length = usb_usertransfer_length;
    if (length > EP_BUFFERSIZE)
        length = EP_BUFFERSIZE;

    length = usb_vendor_in((unsigned char*) ep0in_buffer, length);

    usb_usertransfer_length -= length;

    ep[0].in.cnt = length;
    if (ep[0].in.stat & 0x40)
        ep[0].in.stat = 0x88;
    else
        ep[0].in.stat = 0xC8;

    if (length < EP_BUFFERSIZE) {
        usb_usertransfer = 0;
    }
}




void usb_ep1_arm() {
    
    //ep[1].in.cnt = EP_BULK_BUFFERSIZE;

    if (ep[1].in.stat & 0x40)
        ep[1].in.stat = 0x88;
    else
        ep[1].in.stat = 0xC8;    
}

unsigned char usb_ep1_isAvailable() {
    if (!configured) return 0;
    if (ep[1].in.stat & 0x80) return 0;
    
    return 1;    
}

volatile unsigned char * usb_ep1_getBuffer() {
    return ep1in_buffer;
}

/**
 * @brief Do USB stack processing
 */
void usb_process() {

    if (UIRbits.TRNIF) {
        // complete interrupt

        if (USTAT == USTAT_EP0_OUT) {

            // out/setup

            if (((ep[0].out.stat >> 2) & 0x0F) == USB_PID_SETUP) {
                // setup token

                usb_usertransfer = 0;

                // stop pending/old/prepared IN transactions
                ep[0].in.stat = 0;
                usb_sendpending = 0;

                if ((ep0out_buffer[0] & USBRQ_TYPE_MASK) == USBRQ_TYPE_STANDARD) {

                    switch (ep0out_buffer[1]) {
                        case REQUEST_GET_DESCRIPTOR:
                            if (!usb_handleDescriptorRequest(ep0out_buffer[3], ep0out_buffer[2], (ep0out_buffer[7] << 8) | ep0out_buffer[6])) {
                                ep[0].in.cnt = 0;
                                ep[0].in.stat = 0xCC; // Stall
                            }
                            break;
                        case REQUEST_SET_ADDRESS:

                            usb_setaddress = ep0out_buffer[2];

                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;

                            break;

                        case REQUEST_SET_CONFIGURATION:

                            usb_config = ep0out_buffer[2];
                            configured = 1;
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_CONFIGURATION:

                            ep0in_buffer[0] = usb_config;
                            ep[0].in.cnt = 1;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_INTERFACE:

                            ep0in_buffer[0] = 0;
                            ep[0].in.cnt = 1;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_STATUS:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                ep0in_buffer[0] = usb_ep0status[0];
                                ep0in_buffer[1] = usb_ep0status[1];
                            } else {
                                ep0in_buffer[0] = 0;
                                ep0in_buffer[1] = 0;
                            }
                            ep[0].in.cnt = 2;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SET_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                if (ep0out_buffer[2] == 0x00) // HALT
                                    usb_ep0status[0] = 1;
                            }
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_CLEAR_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                usb_ep0status[0] = 0;
                            }
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SYNCH_FRAME:
                            ep0in_buffer[0] = 0;
                            ep0in_buffer[1] = 0;
                            ep[0].in.cnt = 2;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SET_INTERFACE:
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        default:
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xCC; // stall
                            break;

                    }
                } else {

                    // not standard type -> class or vendor

                    if ((ep0out_buffer[0] & 0x80) == 0x80) {
                        // vendor or class control in

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);

                        if (len == 0xff) {
                            usb_usertransfer_length = (ep0out_buffer[7] << 8) | ep0out_buffer[6];
                            usb_usertransfer = 1;
                            ut_in();
                        } else {
                            ep[0].in.cnt = len;
                            ep[0].in.stat = 0xC8;
                        }


                    } else {
                        // vendor or class control out

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);
                        if (len == 0xff) {
                            usb_usertransfer_length = (ep0out_buffer[7] << 8) | ep0out_buffer[6];
                            usb_usertransfer = 2;
                        } else {
                            // TODO: what to do here? is zero package correct?
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                        }
                    }
                }

            } else {
                // USB_PID_OUT

                // stop pending/old/prepared IN transactions
                ep[0].in.stat = 0;
                usb_sendpending = 0;

                if (usb_usertransfer == 2) {

                    unsigned char ack = usb_vendor_out((unsigned char*) ep0out_buffer, ep[0].out.cnt);

                    if (ack) {
                        ep[0].in.cnt = 0;
                        ep[0].in.stat = 0xC8;
                    } else {
                        ep[0].in.cnt = 0;
                        ep[0].in.stat = 0xCC; // Stall
                    }
                }

            }

            ep[0].out.cnt = EP_BUFFERSIZE;
            ep[0].out.stat = 0x80;


        } else if (USTAT == USTAT_EP0_IN) {

            // check if set address command is pending
            if (usb_setaddress > 0) {
                UADDR = usb_setaddress;
                usb_setaddress = 0;
            }

            if (usb_usertransfer == 1) {
                ut_in();
            } else {
                usb_sendProcess();
            }
        }

        UCONbits.PKTDIS = 0;
        UIRbits.TRNIF = 0;
    }
}
